open Sml;;

let b = Buffer.create 100;;

let rec tok = parser
 | [< ''<'; t = lex_tag >] -> Some t
 | [< ' (' ' | '\n' | '\r' | '\t' as c) >] -> Some (Space c)
 | [< ' ('\033'..';' | '=' | '?'..'\126' as c); s >] ->
     Buffer.add_char b c;
     Some (lex_text s)
 | [< >] -> None

and lex_text = parser
 | [< ' ('\033'..';' | '=' | '?'..'\126' as c); s >] ->
     Buffer.add_char b c;
     lex_text s
 | [<>] ->
     let t = Buffer.contents b in
     Buffer.reset b;
     Text t

and lex_tag_type = parser
 | [< ''U'; ''>' >] -> U
 | [< ''S'; ''>' >] -> S
 | [< ''E'; ''M'; ''>' >] -> EM
 | [< ''I'; ''>' >] -> I
 | [< ''B'; ''>' >] -> B
 | [< ''T'; ''T'; ''>' >] -> TT
 | [< ''P'; ''L'; ''>' >] -> PL
 | [< ''r'; ''>' >] -> Color Red
 | [< ''g'; ''>' >] -> Color Green
 | [< ''b'; ''>' >] -> Color Blue
 | [< ''c'; ''>' >] -> Color Cyan
 | [< ''m'; ''>' >] -> Color Magenta
 | [< ''y'; ''>' >] -> Color Yellow
 | [< ''k'; ''>' >] -> Color Black
 | [< ''w'; ''>' >] -> Color White
 | [< ''0'..'9' as c; ''>' >] -> SZ (Char.code c - Char.code '0')

and lex_tag = parser
 | [< ''/'; s >] -> Close (lex_tag_type s)
 | [< ty = lex_tag_type >] -> Open ty
;;

let lex ic =
  let str = Stream.of_channel ic in
  let rec lx () =
    match tok str with
    | None -> []
    | Some t -> t :: (lx ())
  in lx ()
;;

let llex ic =
  let str = Stream.of_channel ic in
  let rec lx () =
    match tok str with
    | None -> Lnil
    | Some t -> Lcons (t, lazy (lx ()))
  in lx ()
;;

