open Sml;;

let string_of_color = function
 | Red -> "r"
 | Green -> "g"
 | Blue -> "b"
 | Cyan -> "c"
 | Magenta -> "m"
 | Yellow -> "y"
 | Black -> "k"
 | White -> "w"
 | Initial -> assert false
;;

let string_of_tag = function
 | U -> "U"
 | S -> "S"
 | EM -> "EM"
 | I -> "I"
 | B -> "B"
 | TT -> "TT"
 | PL -> "PL"
 | SZ int -> string_of_int int
 | Color color -> string_of_color color
;;

let rand_color () = match Random.int 8 with
  | 0 -> Red
  | 1 -> Green 
  | 2 -> Blue 
  | 3 -> Cyan 
  | 4 -> Magenta 
  | 5 -> Yellow 
  | 6 -> Black 
  | _ -> White;;

let rand_size () = Random.int 10;;

let rand_sml_text sz =
 let rec rand_sml_char () =
  (* '033'..'059' | '061' | '?'..'~'
     (i.e. '!'..';' | '=' | '063'..'\126')
     i.e. 126 - 33 + 1 - 2 = 92 chars *)
  let c = 33 + Random.int 94 in
  if c = 60 || c = 62 then rand_sml_char () else
  char_of_int c in

 let t = String.create sz in
 for i = 0 to sz - 1 do t.[i] <- rand_sml_char () done;
 t;;

let rand_text () =
 let sz = 1 + Random.int 10 in
 rand_sml_text sz;;

let rand_tag () =
 match Random.int 9 with
 | 0 -> U
 | 1 -> S
 | 2 -> EM
 | 3 -> I
 | 4 -> B
 | 5 -> TT
 | 6 -> PL
 | 7 -> SZ (rand_size ())
 | 8 -> Color (rand_color ())
 | _ -> assert false;;

let rand_space () =
 match Random.int 4 with
 | 0 -> ' '
 | 1 -> '\n'
 | 2 -> '\n'
 | 3 -> ' '
 | _ -> assert false;;
 
let rand_lexem () =
  match Random.int 4 with
  | 0 -> Text (rand_text ())
  | 1 -> Space (rand_space ())
  | 2 -> Open (rand_tag ())
  | 3 -> Close (rand_tag ())
  | _ -> assert false;;

open Printf;;

let cur_col = ref 0;;

let print l =
  begin match l with
  | Text s ->
      printf "%s" s;
      cur_col := !cur_col + String.length s;
  | Space '\t' ->
      printf "\t";
      cur_col := !cur_col + 8;
  | Space '\n' ->
      printf "\n";
      cur_col := 0;
  | Space c ->
      printf "%c" c;
      cur_col := !cur_col + 1;
  | Open tag ->
      printf "<%s>" (string_of_tag tag);
      cur_col := !cur_col + 2 + String.length (string_of_tag tag);
  | Close tag ->
      printf "</%s>" (string_of_tag tag);
      cur_col := !cur_col + 3 + String.length (string_of_tag tag);
  end;
  if !cur_col > 68 then begin
    printf "\n";
    cur_col := 0;
  end;
;;

let rand_sml limit =

 let rec reopen to_reopen =
   match to_reopen with
   | [] -> ()
   | t :: rest -> print (Open t); reopen rest in

 let rec close_all = function
   | [] -> ()
   | t :: rest -> print (Close t); close_all rest in

 let rec rand stack count =
   if count >= limit then close_all stack else
   let lexem = rand_lexem () in
   (*print Format.std_formatter lexem; Format.print_newline();*)
   match lexem with
   | Open t -> print lexem; rand (t :: stack) (count + 1)
   | Close t ->
      if List.mem t stack then
       let rec loop to_reopen = function
         | [] ->
             print (Open t);
             print (Close t);
             reopen to_reopen;
             rand stack (count + List.length to_reopen + 2)
         | t' :: rest ->
             if t = t' then begin
               print lexem;
               rand rest (count + 1)
             end else begin
               print (Close t');
               loop (t' :: to_reopen) rest
             end
       in
       loop [] stack else
      (* Once out of twice consider this close as an open,
         otherwise simply ignore it. *)
      if Random.int 2 = 1
      then begin
        print (Open t);
        rand (t :: stack) (count + 1);
      end else begin
        rand stack count
      end
   | _ -> print lexem; rand stack (count + 1)
   in
   rand [] 0;
   []
;;
